from .util import doesMatchPattern, check_key_in_dict, get_roi
from .ReadData import REIXS
from .edges import EdgeDict
from .xeol import *
from .simplemath import apply_offset, grid_data_mesh
import warnings
import numpy as np
from .parser import math_stream


def loadMeshScans(basedir, file, x_stream, y_stream, z_stream, *args, norm=True, is_XAS=False, xoffset=None, xcoffset=None, yoffset=None, ycoffset=None, background=None):
    """Internal function to generate scatter plots for (x,y,z) SCA data"""

    # Define special streams one might want to consider
    special_streams = ['TEY', 'TFY', 'PFY', 'iPFY', 'XES', 'rXES', 'specPFY',
                       'XRF', 'rXRF', 'XEOL', 'rXEOL', 'POY', 'TOY', 'EY']  # all special inputs
    # Append also to XAS_Streams variable, if the special stream is already normalized to i0 (incident flux)
    XAS_streams = ['TEY', 'TFY', 'PFY', 'iPFY', 'specPFY', 'POY',
                   'TOY', 'rXES', 'rXRF', 'rXEOL']  # All that are normalized to mesh

    # Note that the data dict only gets populated locally until the appropriate
    # singular y stream is return -- Assignment of y_stream happens after evaluation
    # of mathemtatical expression

    def get_z_data(z_stream, data, arg, background, REIXSObj):
        # Get the z-data stream
        # Check if the requested stream is a special stream
        if doesMatchPattern(z_stream, special_streams):
            # CAUTION: Order matters as doesmatchpattern also returns TRUE if pattern in instance
            # is recogniced --> shortest pattern to check last

            # Check which special stream is requested and strip the arguments for ROI if applicable
            if doesMatchPattern(z_stream, ['rXES']):
                roi = z_stream.lstrip("rXES[").rstrip("]")

                roi_low, roi_high = get_roi(roi)

                return data[arg].rXES(roi_low, roi_high)

            elif z_stream == 'XES':
                return data[arg].XES()

            elif doesMatchPattern(z_stream, ['rXRF']):
                roi = z_stream.lstrip("rXRF[").rstrip("]")
                roi_low, roi_high = get_roi(roi)

                return data[arg].rXRF(roi_low, roi_high)

            elif z_stream == 'XRF':
                return data[arg].XRF()

            elif z_stream == 'TEY':
                return data[arg].TEY

            elif z_stream == 'TFY':
                return data[arg].TFY()

            elif doesMatchPattern(z_stream, ["iPFY"]):
                roi = z_stream.lstrip("iPFY[").rstrip("]")

                if check_key_in_dict(roi, EdgeDict):
                    return data[arg].iPFY(iPFY_edge=roi)

                else:
                    roi_low, roi_high = get_roi(roi)
                    return data[arg].iPFY(iSDDLowerBound=roi_low, iSDDUpperBound=roi_high)

            elif doesMatchPattern(z_stream, ["specPFY"]):
                roi = z_stream.lstrip("specPFY[").rstrip("]")
                roi_low, roi_high = get_roi(roi)
                return data[arg].specPFY(roi_low, roi_high)

            elif doesMatchPattern(z_stream, ["PFY"]):
                roi = z_stream.lstrip("PFY[").rstrip("]")

                if check_key_in_dict(roi, EdgeDict):
                    return data[arg].PFY(PFY_edge=roi)

                else:
                    roi_low, roi_high = get_roi(roi)
                    return data[arg].PFY(SDDLowerBound=roi_low, SDDUpperBound=roi_high)

            elif z_stream == 'XEOL':
                return xeol_spec(data, arg, REIXSobj, background_scan=background)

            elif doesMatchPattern(z_stream, ["rXEOL"]):
                roi = z_stream.lstrip("rXEOL[").rstrip("]")
                roi_low, roi_high = get_roi(roi)

                return rxeol_spec(data, arg, REIXSobj, roi_low, roi_high, background_scan=background)

            elif z_stream == 'TOY':
                return toy_spec(data, arg, REIXSobj, background_scan=background)

            elif doesMatchPattern(z_stream, ["POY"]):
                roi = z_stream.lstrip("POY[").rstrip("]")
                roi_low, roi_high = get_roi(roi)

                return poy_spec(data, arg, REIXSobj, roi_low, roi_high, background_scan=background)

            elif z_stream == 'EY':
                return data[arg].sample_current

            else:
                try:
                    # Else, load from pandas SCA data frame
                    return get_sca_data(z_stream, data, arg)
                except:
                    raise UserWarning("Special Stream not defined.")

        else:
            return get_sca_data(z_stream, data, arg)

    def get_sca_data(stream, data, arg):
        # Read in stream with long name or utilize mnemonic dict
        try:
            return np.array(data[arg].sca_data[stream])
        except:
            return np.array(data[arg].sca_data[data[arg].mnemonic2name[stream]])

    def get_x_data(x_stream, data, arg, background, REIXSObj):
        # Get the x-data.
        if x_stream == "Mono Energy":
            return data[arg].mono_energy

        elif x_stream == "MCP Energy":
            return data[arg].mcp_energy

        elif x_stream == "SDD Energy":
            return data[arg].sdd_energy

        elif x_stream == "XEOL Energy":
            return data[arg].xeol_energy

        else:
            return get_sca_data(x_stream, data, arg)

    def get_y_data(y_stream, data, arg, background, REIXSObj):
        # Get the y data
        if y_stream == "Mono Energy":
            return data[arg].mono_energy

        elif y_stream == "MCP Energy":
            return data[arg].mcp_energy

        elif y_stream == "SDD Energy":
            return data[arg].sdd_energy

        elif y_stream == "XEOL Energy":
            return data[arg].xeol_energy

        else:
            return get_sca_data(y_stream, data, arg)

    # Generate dictionary to store data with REIXS objects
    data = dict()
    REIXSobj = REIXS(basedir, file)
    for arg in args:
        # Load scans to dict
        data[arg] = REIXSobj.Scan(arg)

        # Assign the calculated result to the y_stream of the object in data dict
        data[arg].y_data = math_stream(y_stream, data, arg, get_y_data)
        data[arg].y_data = apply_offset(data[arg].y_data, yoffset, ycoffset)

        # Assign the calculated result to the x_stream of the object in data dict
        data[arg].x_data = math_stream(x_stream, data, arg, get_x_data)
        data[arg].x_data = apply_offset(data[arg].x_data, xoffset, xcoffset)

        # Aplly simple math to x-stream
        data[arg].z_data = math_stream(z_stream, data, arg, get_z_data, XAS_streams=XAS_streams,
                                       is_XAS=is_XAS, background=background, REIXSObj=REIXSobj)

        # Normalize if requested
        if norm == True:
            data[arg].z_data = np.interp(
                data[arg].z_data, (data[arg].z_data.min(), data[arg].z_data.max()), (0, 1))

        xmin, xmax, ymin, ymax, xedge, yedge, new_z, zmin, zmax = grid_data_mesh(data[arg].x_data,data[arg].y_data,data[arg].z_data)
        data[arg].xmin = xmin
        data[arg].xmax = xmax
        data[arg].ymin = ymin
        data[arg].ymax = ymax
        data[arg].xedge = xedge
        data[arg].yedge = yedge
        data[arg].new_z = new_z
        data[arg].zmin = zmin
        data[arg].zmax = zmax

    return data



import numpy as np
from scipy.interpolate import interp2d

def grid_data(mcpRSXS_scales, mcpRSXS_scatters, img, x_low=None, x_high=None, y_low=None, y_high=None, new_x=None, new_y=None):

    mcpRSXS_scale = mcpRSXS_scales[img]
    mcpRSXS_scatter = mcpRSXS_scatters[img]

    if x_low != None:
        if x_low < mcpRSXS_scale[0, :].min():
            xmin = mcpRSXS_scale[0, :].min()
        else:
            xmin = x_low
    else:
        xmin = mcpRSXS_scale[0, :].min()

    if x_high != None:
        if x_high < mcpRSXS_scale[0, :].max():
            xmax = x_high
        else:
            xmax = mcpRSXS_scale[0, :].max()
    else:
        xmax = mcpRSXS_scale[0, :].max()

    if y_low != None:
        if y_low < mcpRSXS_scale[1, :].min():
            ymin = mcpRSXS_scale[1, :].min()
        else:
            ymin = y_low
    else:
        ymin = mcpRSXS_scale[1, :].min()

    if y_high != None:
        if y_high < mcpRSXS_scale[1, :].max():
            ymax = y_high
        else:
            ymax = mcpRSXS_scale[1, :].max()
    else:
        ymax = mcpRSXS_scale[1, :].max()

    xpoints = int(
        np.abs(np.ceil((xmax-xmin)/np.abs(np.diff(mcpRSXS_scale[0, :])).min())))
    ypoints = int(
        np.abs(np.ceil((ymax-ymin)/np.abs(np.diff(mcpRSXS_scale[1, :])).min())))

    f = interp2d(mcpRSXS_scale[0], mcpRSXS_scale[1], mcpRSXS_scatter)

    if type(new_x) == type(None):
        new_x = np.linspace(xmin, xmax, xpoints)
    if type(new_y) == type(None):
        new_y = np.linspace(ymin, ymax, ypoints)

    new_z = f(new_x, new_y)
    new_z = np.transpose(new_z)

    return xmin, xmax, ymin, ymax, new_x, new_y, new_z


def img_to_sca(mcpRSXS_scales, mcpRSXS_scatters, img, x_low=None, x_high=None, y_low=None, y_high=None, axis=0, new_x=None, new_y=None):

    xmin, xmax, ymin, ymax, new_x, new_y, new_z = grid_data(
        mcpRSXS_scales, mcpRSXS_scatters, img, x_low, x_high, y_low, y_high, new_x, new_y)

    newz_sum = np.sum(new_z, axis=axis)

    if axis == 0:
        caxis = new_y
    elif axis == 1:
        caxis = new_x
    else:
        raise TypeError("Invalid axis")

    return caxis, newz_sum


def grid_stack(mcpRSXS_scales, mcpRSXS_scatters):

    z_stack = dict()
    x_min = dict()
    x_max = dict()
    y_min = dict()
    y_max = dict()

    for i in range(len(mcpRSXS_scatters)):
        xmin, xmax, ymin, ymax, new_x, new_y, new_z = grid_data(
            mcpRSXS_scales, mcpRSXS_scatters, i)

        z_stack[i] = new_z
        x_min[i] = xmin
        x_max[i] = xmax
        y_min[i] = ymin
        y_max[i] = ymax

    return z_stack, x_min, x_max, y_min, y_max


def check_array_ordering(mcpRSXS_scales, dictkey):
    if mcpRSXS_scales[dictkey][0][0] < mcpRSXS_scales[dictkey][0][1]:
        sorter_0 = 'asc'
    else:
        sorter_0 = 'desc'

    if mcpRSXS_scales[dictkey][1][0] < mcpRSXS_scales[dictkey][1][1]:
        sorter_1 = 'asc'
    else:
        sorter_1 = 'desc'

    return sorter_0, sorter_1


def get_interpolation_boundaries(mcpRSXS_scales):

    sorter_0_vals_min = []
    sorter_0_vals_max = []
    sorter_1_vals_min = []
    sorter_1_vals_max = []
    min_diff_0_vals = []
    min_diff_1_vals = []

    sorter_0, sorter_1 = check_array_ordering(mcpRSXS_scales, 0)

    for k, v in mcpRSXS_scales.items():
        sorter_0_vals_min.append(v[0, 0])
        sorter_1_vals_min.append(v[1, 0])

        sorter_0_vals_max.append(v[0, -1])
        sorter_1_vals_max.append(v[1, -1])

        min_diff_0_vals.append(np.abs(np.diff(v[0])).min())
        min_diff_1_vals.append(np.abs(np.diff(v[1])).min())

        min_diff_0 = min(min_diff_0_vals)
        min_diff_1 = min(min_diff_1_vals)

        if sorter_0 == 'asc':
            sorter_0_min = max(sorter_0_vals_min)
            sorter_0_max = min(sorter_0_vals_max)
        else:
            sorter_0_min = min(sorter_0_vals_min)
            sorter_0_max = max(sorter_0_vals_max)

        if sorter_1 == 'asc':
            sorter_1_min = max(sorter_1_vals_min)
            sorter_1_max = min(sorter_1_vals_max)

        else:
            sorter_1_min = min(sorter_1_vals_min)
            sorter_1_max = max(sorter_1_vals_max)

    return sorter_0_min, sorter_0_max, sorter_1_min, sorter_1_max, min_diff_0, min_diff_1


def stack_to_mca(mcpRSXS_scales, mcpRSXS_scatters, x_low=None, x_high=None, y_low=None, y_high=None, axis=0):

    sorter_0_min, sorter_0_max, sorter_1_min, sorter_1_max, min_diff_0, min_diff_1 = get_interpolation_boundaries(
        mcpRSXS_scales)

    if x_low != None:
        if x_low < sorter_0_min:
            xmin = sorter_0_min
        else:
            xmin = x_low
    else:
        xmin = sorter_0_min

    if x_high != None:
        if x_high < sorter_0_max:
            xmax = x_high
        else:
            xmax = sorter_0_max
    else:
        xmax = sorter_0_max

    if y_low != None:
        if y_low < sorter_1_min:
            ymin = sorter_1_min
        else:
            ymin = y_low
    else:
        ymin = sorter_1_min

    if y_high != None:
        if y_high < sorter_1_max:
            ymax = y_high
        else:
            ymax = sorter_1_max
    else:
        ymax = sorter_1_max

    xpoints = int(np.abs(np.ceil((xmax-xmin)/min_diff_0)))
    ypoints = int(np.abs(np.ceil((ymax-ymin)/min_diff_1)))

    new_x = np.linspace(xmin, xmax, xpoints)
    new_y = np.linspace(ymin, ymax, ypoints)

    mca = []
    for i in range(len(mcpRSXS_scatters)):
        caxis, newz_sum = img_to_sca(
            mcpRSXS_scales, mcpRSXS_scatters, i, xmin, xmax, ymin, ymax, axis, new_x, new_y)
        mca.append(newz_sum)

    if axis == 0:
        new_x = np.arange(0, len(mcpRSXS_scales))
        image = np.transpose(np.stack(mca))

    elif axis == 1:
        new_y = np.arange(0, len(mcpRSXS_scatters))
        image = np.stack(mca)
    else:
        raise TypeError("Invalid axis.")

    return new_x, new_y, image